/***************************************************************************
    smb4kpasswordhandler  -  This class handles the passwords for Smb4K.
                             -------------------
    begin                : So Jan 16 2005
    copyright            : (C) 2005-2008 by Alexander Reinholdt
    email                : dustpuppy@users.berlios.de
 ***************************************************************************/

/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful, but   *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,   *
 *   MA  02111-1307 USA                                                    *
 ***************************************************************************/


// Qt includes
#include <QGridLayout>
#include <QLabel>
#include <QDir>
#include <QFile>
#include <QStringList>
#include <QDesktopWidget>
#include <QDBusConnection>
#include <QDBusConnectionInterface>

// KDE includes
#include <klocale.h>
#include <kstandarddirs.h>
#include <kiconloader.h>
#include <klineedit.h>
#include <kcombobox.h>
#include <kdebug.h>
#include <kapplication.h>
#include <ktoolinvocation.h>
#include <klauncher_iface.h>
#include <kglobal.h>

#ifdef __FreeBSD__
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#endif

// application specific includes
#include <smb4kpasswordhandler.h>
#include <smb4kdefs.h>
#include <smb4kcoremessage.h>
#include <smb4kauthinfo.h>
#include <smb4ksettings.h>
#include <smb4kshare.h>

class Smb4KPasswordHandlerPrivate
{
  public:
    /**
     * The Smb4KPasswordHandler instance
     */
    Smb4KPasswordHandler instance;
};


K_GLOBAL_STATIC( Smb4KPasswordHandlerPrivate, m_priv );


Smb4KPasswordHandler::Smb4KPasswordHandler()
: QObject()
{
#ifdef __FreeBSD__
  m_proc = new KProcess( this );
  m_process_error = (QProcess::ProcessError)(-1);

  connect( m_proc, SIGNAL( finished( int, QProcess::ExitStatus ) ),
           this,   SLOT( slotProcessFinished( int, QProcess::ExitStatus ) ) );

  connect( m_proc, SIGNAL( error( QProcess::ProcessError ) ),
           this,   SLOT( slotProcessError( QProcess::ProcessError ) ) );

#endif

  m_auth = NULL;
  m_dlg = NULL;
  m_wallet = NULL;
  m_temp_auth = NULL;

  m_wallet_support_disabled = false;
  m_can_read_auth_info = false;
}


Smb4KPasswordHandler::~Smb4KPasswordHandler()
{
  while ( !m_auth_list.isEmpty() )
  {
    delete m_auth_list.takeFirst();
  }

//   delete m_wallet;
}


Smb4KPasswordHandler *Smb4KPasswordHandler::self()
{
  return &m_priv->instance;
}



void Smb4KPasswordHandler::init( QWidget *parent, bool async )
{
  if ( Smb4KSettings::useWallet() )
  {
    if ( !m_wallet_support_disabled )
    {
      if ( !walletIsOpen() )
      {
        // Start KLauncher, so that we can access the wallet
        // (otherwise we get a problem when KDE is starting up).
        if ( !QDBusConnection::sessionBus().interface()->isServiceRegistered( "org.kde.klauncher" ).value() )
        {
          KToolInvocation::self()->klauncher()->autoStart();
        }
        else
        {
          // Do nothing
        }

        // Attempt to open the wallet.
        WId parentId = 0;

        if ( !parent )
        {
          if ( kapp )
          {
            if ( kapp->activeWindow() )
            {
              parentId = kapp->activeWindow()->winId();
            }
            else
            {
              parentId = kapp->desktop()->winId();
            }
          }
          else
          {
            // Do nothing
          }
        }
        else
        {
          parentId = parent->winId();
        }

        if ( async )
        {
          m_wallet = KWallet::Wallet::openWallet( KWallet::Wallet::NetworkWallet(),
                     parentId, KWallet::Wallet::Asynchronous );

          connect( m_wallet, SIGNAL( walletOpened( bool ) ),
                   this,     SLOT( slotWalletOpened( bool ) ) );
        }
        else
        {
          m_wallet = KWallet::Wallet::openWallet( KWallet::Wallet::NetworkWallet(),
                     parentId, KWallet::Wallet::Synchronous );

          if ( m_wallet )
          {
            if ( !m_wallet->hasFolder( "Smb4K" ) )
            {
              m_wallet->createFolder( "Smb4K" );
              m_wallet->setFolder( "Smb4K" );
            }
            else
            {
              m_wallet->setFolder( "Smb4K" );
              convert_old_entries();
            }

            m_wallet_support_disabled = false;
          }
          else
          {
            Smb4KCoreMessage::information( INFO_OPENING_WALLET_FAILED, KWallet::Wallet::NetworkWallet(), QString() );

            delete m_wallet;
            m_wallet = NULL;

            m_wallet_support_disabled = true;
          }

          m_can_read_auth_info = true;
          emit readyReadAuthInfo();
        }
      }
      else
      {
        // Convert old entries.
        convert_old_entries();

        m_can_read_auth_info = true;
        emit readyReadAuthInfo();
      }
    }
    else
    {
      if ( m_wallet )
      {
        // Close the wallet (if possible) and delete it. We
        // will not use it.
        KWallet::Wallet::closeWallet( KWallet::Wallet::NetworkWallet(), false );

        delete m_wallet;
        m_wallet = NULL;
      }
      else
      {
        // Do nothing
      }

      m_can_read_auth_info = true;
      emit readyReadAuthInfo();
    }
  }
  else
  {
    if ( m_wallet )
    {
      // Close the wallet (if possible) and delete it. We
      // will not use it.
      KWallet::Wallet::closeWallet( KWallet::Wallet::NetworkWallet(), false );

      delete m_wallet;
      m_wallet = NULL;
    }
    else
    {
      // Do nothing
    }

    m_can_read_auth_info = true;
    emit readyReadAuthInfo();
  }
}


void Smb4KPasswordHandler::convert_old_entries()
{
  // Convert old entries (Smb4K 0.9.2 and earlier) to the
  // new map based format.
  if ( !m_wallet->entryList().isEmpty() )
  {
    QStringList entries = m_wallet->entryList();

    // Since in the old format the keys contained a string like
    // this: HOST:USER, we only need to test whether the first
    // key contains a ":". If that's the case, we need to convert
    // the entries, otherwise we don't.
    if ( entries.first().contains( ":" ) )
    {
      for ( QStringList::Iterator it = entries.begin(); it != entries.end(); ++it )
      {
        // Get the password.
        QString pass;
        m_wallet->readPassword( *it, pass );

        if ( (*it).startsWith( "DEFAULT:" ) )
        {
          // Convert the data to the new format.
          QMap<QString,QString> map;
          map["Login"]     = (*it).section( ":", 1, 1 );
          map["Password"]  = pass;

          m_wallet->writeMap( "DEFAULT_LOGIN", map );
        }
        else
        {
          // Convert the data to the new format.
          QMap<QString,QString> map;
          map["Login"]     = (*it).section( ":", 3, 3 );
          map["Password"]  = pass;

          if ( QString::compare( (*it).section( ":", 0, 0 ), "*" ) != 0 )
          {
            map["Workgroup"] = (*it).section( ":", 0, 0 ).toUpper();
          }

          if ( QString::compare( (*it).section( ":", 2, 2 ), "*" ) == 0 )
          {
            m_wallet->writeMap( (*it).section( ":", 1, 1 ).toUpper(), map );
          }
          else
          {
            m_wallet->writeMap( "//"+(*it).section( ":", 1, 1 ).toUpper()+
                                "/"+(*it).section( ":", 2, 2 ).toUpper(), map );
          }
        }

        // Delete the old entry.
        m_wallet->removeEntry( *it );
      }
    }
    else
    {
      // Do nothing.
    }
  }
  else
  {
    // Do nothing
  }
}


bool Smb4KPasswordHandler::askpass( const QString &workgroup, const QString &host, const QString &share, Description desc, QWidget *parent )
{
  if ( !parent )
  {
    if ( kapp )
    {
      if ( kapp->activeWindow() )
      {
        parent = kapp->activeWindow();
      }
      else
      {
        parent = kapp->desktop();
      }
    }
    else
    {
      parent = 0;
    }
  }
  else
  {
    // Do nothing
  }

  // Initialize.
  init( parent, false );

  // m_auth is NULL:
  m_auth = new Smb4KAuthInfo( workgroup, host, share );

  readAuth( m_auth );

  m_dlg = new KDialog( parent );
  m_dlg->setCaption( i18n( "Authentication" ) );
  m_dlg->setButtons( KDialog::Ok | KDialog::Cancel );
  m_dlg->setDefaultButton( KDialog::Ok );

  QWidget *frame = new QWidget( m_dlg );
  QGridLayout *layout = new QGridLayout( frame );
  layout->setSpacing( 5 );
  layout->setMargin( 0 );

  m_dlg->setMainWidget( frame );

  QLabel *pixmap_label = new QLabel( frame );
  pixmap_label->setPixmap( DesktopIcon( "dialog-password" ) );
  pixmap_label->adjustSize();

  layout->addWidget( pixmap_label, 0, 0, Qt::AlignCenter );

  QString message;

  switch ( desc )
  {
    case NewData:
    {
      break;
    }
    case AccessDenied:
    {
      message = i18n( "The access was denied. " );
      break;
    }
    case BadPassword:
    {
      message = i18n( "The password is not correct. " );
      break;
    }
    case PermissionDenied:
    {
      message = i18n( "The permission was denied. " );
      break;
    }
    case AuthenticationError:
    {
      message = i18n( "An authentication error occurred. " );
      break;
    }
    case LogonFailure:
    {
      message = i18n( "The logon failed. " );
      break;
    }
    default:
    {
      break;
    }
  }

  if ( m_auth->share().trimmed().isEmpty() )
  {
    message.append( i18n( "Please enter authentication data for server %1." ).arg( m_auth->host() ) );
  }
  else
  {
    message.append( i18n( "Please enter authentication data for share %1." ).arg( "//"+m_auth->host()+"/"+m_auth->share() ) );
  }

  QLabel *message_label = new QLabel( frame );
  message_label->setText( message.trimmed() );
  message_label->setTextFormat( Qt::RichText );
  message_label->setWordWrap( true );

  layout->addWidget( message_label, 0, 1, 0 );

  QLabel *user_label = new QLabel( i18n( "User:" ), frame );
  layout->addWidget( user_label, 1, 0, 0 );

  KLineEdit *user_edit = NULL;
  KComboBox *user_combo = NULL;

  if ( QString::compare( m_auth->share(), "homes" ) != 0 )
  {
    user_edit = new KLineEdit( frame );
    user_edit->setObjectName( "AskPassUserEdit" );
    user_edit->setMinimumWidth( 200 );
    layout->addWidget( user_edit, 1, 1, 0 );
  }
  else
  {
    user_combo = new KComboBox( frame );
    user_combo->setObjectName( "AskPassUserCombo" );
    user_combo->setEditable( true );
    user_combo->setMinimumWidth( 200 );
    layout->addWidget( user_combo, 1, 1, 0 );
  }

  QLabel *password_label = new QLabel( i18n( "Password:" ), frame );
  layout->addWidget( password_label, 2, 0, 0 );

  KLineEdit *pass_edit = new KLineEdit( frame );
  pass_edit->setObjectName( "AskPassPasswordEdit" );
  pass_edit->setEchoMode( KLineEdit::Password );
  layout->addWidget( pass_edit, 2, 1, 0 );

  m_dlg->setMainWidget( frame );
  m_dlg->setFixedSize( 350, m_dlg->sizeHint().height() );
  m_dlg->enableButtonOk( false );

  // Since we have to allow empty passwords, we will only connect
  // the edit line for the user to enable/disable the OK button.
  if ( user_edit )
  {
    connect( user_edit, SIGNAL( textChanged( const QString & ) ),
             this,      SLOT( slotEnableOKButton( const QString& ) ) );
  }
  else
  {
    connect( user_combo, SIGNAL( textChanged( const QString & ) ),
             this,       SLOT( slotEnableOKButton( const QString& ) ) );
  }

  // Process the authentication data:

  if ( QString::compare( share, "homes" ) != 0 )
  {
    user_edit->setText( m_auth->user() );
    pass_edit->setText( m_auth->password() );

    if ( m_auth->user().isEmpty() )
    {
      user_edit->setFocus();
    }
    else
    {
      pass_edit->setFocus();
    }
  }
  else
  {
    Smb4KShare homes_share;
    homes_share.setWorkgroup( workgroup );
    homes_share.setHost( host );
    homes_share.setName( share );

    Smb4KHomesSharesHandler::self()->setHomesUsers( &homes_share );

    if ( !homes_share.homesUsers().isEmpty() )
    {
      user_combo->insertItems( 1, homes_share.homesUsers() );
    }

    user_combo->setEditText( QString() );

    connect( user_combo, SIGNAL( activated( const QString & ) ),
             this,       SLOT( slotGetPassword( const QString & ) ) );

    user_combo->setFocus();
  }

  bool ok = false;

  if ( m_dlg->exec() == KDialog::Accepted )
  {
    if ( QString::compare( share, "homes" ) != 0 )
    {
      QString user = user_edit->text();
      QString pass = pass_edit->text();

      m_auth->setUser( user );
      m_auth->setPassword( pass );

      writeAuth( m_auth );
    }
    else
    {
      QString user = user_combo->currentText();
      QString pass = pass_edit->text();

      m_auth->setUser( user );
      m_auth->setPassword( pass );

      writeAuth( m_auth );
    }

    ok = true;
  }

  // Clean up:

  delete m_dlg;
  m_dlg = NULL;

  delete m_auth;
  m_auth = NULL;

  return ok;
}


bool Smb4KPasswordHandler::readAuth( Smb4KAuthInfo *authInfo )
{
  bool success = false;

  if ( m_can_read_auth_info && authInfo )
  {
    if ( walletIsOpen() && Smb4KSettings::useWallet() )
    {
      // Get the authentication information from the wallet.
      // Always prefer either the exact match or the information
      // that was provided for the host.
      QMap<QString,QString> map;

      if ( !authInfo->share().isEmpty() )
      {
        m_wallet->readMap( "//"+authInfo->host().toUpper()+"/"+authInfo->share().toUpper(), map );

        if ( map.isEmpty() )
        {
          m_wallet->readMap( authInfo->host().toUpper(), map );

          if ( map.isEmpty() )
          {
            if ( Smb4KSettings::useDefaultLogin() )
            {
              m_wallet->readMap( "DEFAULT_LOGIN", map );

              if ( map.isEmpty() )
              {
                return success;  // false
              }
              else
              {
                success = true;
              }
            }
            else
            {
              return success;  // false
            }
          }
          else
          {
            // Check that the workgroup is correct.
            if ( map.contains( "Workgroup" ) && !authInfo->workgroup().isEmpty() )
            {
              if ( QString::compare( map["Workgroup"].toUpper(), authInfo->workgroup().toUpper() ) != 0 )
              {
                return success;  // false
              }
              else
              {
                success = true;
              }
            }
            else
            {
              // We cannot check if the workgroup is OK. So, assume that the
              // collected authentication information is the correct one. If it
              // should be wrong, the user has to supply the correct one anyway.
              success = true;
            }
          }
        }
        else
        {
          // Check that the workgroup is correct.
          if ( map.contains( "Workgroup" ) && !authInfo->workgroup().isEmpty() )
          {
            if ( QString::compare( map["Workgroup"].toUpper(), authInfo->workgroup().toUpper() ) != 0 )
            {
              return success;  // false
            }
            else
            {
              success = true;
            }
          }
          else
          {
            // We cannot check if the workgroup is OK. So, assume that the
            // collected authentication information is the correct one. If it
            // should be wrong, the user has to supply the correct one anyway.
            success = true;
          }
        }
      }
      else
      {
        m_wallet->readMap( authInfo->host().toUpper(), map );

        if ( map.isEmpty() )
        {
          if ( Smb4KSettings::useDefaultLogin() )
          {
            m_wallet->readMap( "DEFAULT_LOGIN", map );

            if ( map.isEmpty() )
            {
              return success;  // false
            }
            else
            {
              success = true;
            }
          }
          else
          {
            return success;  // false
          }
        }
        else
        {
          // Check that the workgroup is correct.
          if ( map.contains( "Workgroup" ) && !authInfo->workgroup().isEmpty() )
          {
            if ( QString::compare( map["Workgroup"].toUpper(), authInfo->workgroup().toUpper() ) != 0 )
            {
              return success;  // false
            }
            else
            {
              success = true;
            }
          }
          else
          {
            // We cannot check if the workgroup is OK. So, assume that the
            // collected authentication information is the correct one. If it
            // should be wrong, the user has to supply the correct one anyway.
            success = true;
          }
        }
      }

      // We have authentication information. Put the data into the
      // Smb4KAuthInfo object.
      authInfo->setUser( map["Login"] );
      authInfo->setPassword( map["Password"] );
    }
    else
    {
      if ( !Smb4KSettings::useWallet() && walletIsOpen() )
      {
        delete m_wallet;
        m_wallet = NULL;
      }
      else
      {
        // Do nothing
      }

      if ( Smb4KSettings::rememberLogins() )
      {
        if ( !m_auth_list.isEmpty() )
        {
          for ( int i = 0; i < m_auth_list.size(); ++i )
          {
            if ( !authInfo->share().isEmpty() )
            {
              if ( QString::compare( authInfo->host().toUpper(),
                   m_auth_list.at( i )->host().toUpper() ) == 0 &&
                   QString::compare( authInfo->share().toUpper(),
                   m_auth_list.at( i )->share().toUpper() ) == 0 )
              {
                if ( QString::compare( authInfo->workgroup().toUpper(),
                     m_auth_list.at( i )->workgroup().toUpper() ) == 0 ||
                     m_auth_list.at( i )->workgroup().isEmpty() )
                {
                  // Either the workgroup is OK, or we cannot check if it is OK.
                  // In both cases we process the data at hand.
                  success = true;

                  authInfo->setUser( m_auth_list.at( i )->user() );
                  authInfo->setPassword( m_auth_list.at( i )->password() );

                  // Exact match. Stop here.
                  break;
                }
                else
                {
                  continue;
                }
              }
              else if ( QString::compare( authInfo->host().toUpper(),
                        m_auth_list.at( i )->host().toUpper() ) == 0 )
              {
                if ( QString::compare( authInfo->workgroup().toUpper(),
                     m_auth_list.at( i )->workgroup().toUpper() ) == 0 ||
                     m_auth_list.at( i )->workgroup().isEmpty() )
                {
                  // Either the workgroup is OK, or we cannot check if it is OK.
                  // In both cases we process the data at hand.
                  success = true;

                  authInfo->setUser( m_auth_list.at( i )->user() );
                  authInfo->setPassword( m_auth_list.at( i )->password() );

                  // Because we always prefer the exact match, we keep
                  // on searching.
                  continue;
                }
                else
                {
                  continue;
                }
              }
              else
              {
                continue;
              }
            }
            else
            {
              if ( QString::compare( authInfo->host().toUpper(),
                   m_auth_list.at( i )->host().toUpper() ) == 0 )
              {
                if ( QString::compare( authInfo->workgroup().toUpper(),
                     m_auth_list.at( i )->workgroup().toUpper() ) == 0 ||
                     m_auth_list.at( i )->workgroup().isEmpty() )
                {
                  // Either the workgroup is OK, or we cannot check if it is OK.
                  // In both cases we process the data at hand.
                  success = true;

                  authInfo->setUser( m_auth_list.at( i )->user() );
                  authInfo->setPassword( m_auth_list.at( i )->password() );

                  break;
                }
                else
                {
                  continue;
                }
              }
              else
              {
                continue;
              }
            }
          }
        }
        else
        {
          return success;   // false
        }
      }
      else
      {
        if ( m_temp_auth )
        {
          success = true;

          authInfo->setUser( m_temp_auth->user() );
          authInfo->setPassword( m_temp_auth->password() );

          delete m_temp_auth;
          m_temp_auth = NULL;
        }
      }
    }

#ifdef __FreeBSD__

    writeToSMBConfFile( authInfo );

#endif
  }
  else
  {
    return success;  // false
  }

  return success;
}


bool Smb4KPasswordHandler::writeAuth( Smb4KAuthInfo *authInfo )
{
  bool success = false;

  if ( m_can_read_auth_info && authInfo )
  {
    if ( walletIsOpen() && Smb4KSettings::useWallet() )
    {
      QMap<QString,QString> map;
      map["Login"]       = authInfo->user();
      map["Password"]    = authInfo->password();

      if ( !authInfo->workgroup().isEmpty() )
      {
        map["Workgroup"] = authInfo->workgroup().toUpper();
      }

      if ( !authInfo->share().isEmpty() )
      {
        m_wallet->writeMap( "//"+authInfo->host().toUpper()+"/"+authInfo->share().toUpper(), map );
      }
      else
      {
        m_wallet->writeMap( authInfo->host().toUpper(), map );
      }

      if ( m_wallet->sync() == 0 )
      {
        success = true;
      }
      else
      {
        return success;  // false
      }
    }
    else
    {
      if ( !Smb4KSettings::useWallet() && walletIsOpen() )
      {
        delete m_wallet;
        m_wallet = NULL;
      }
      else
      {
        // Do nothing
      }

      if ( Smb4KSettings::rememberLogins() )
      {
        // Search the temporary list for an entry and modify it. Otherwise
        // add the new auth info.
        Smb4KAuthInfo *known = NULL;

        for ( int i = 0; i < m_auth_list.size(); ++i )
        {
          if ( (m_auth_list.at( i )->workgroup().isEmpty() ||
               QString::compare( m_auth_list.at( i )->workgroup().toUpper(),
               authInfo->workgroup().toUpper() ) == 0) &&
               QString::compare( m_auth_list.at( i )->host().toUpper(),
               authInfo->host().toUpper() ) == 0 &&
               (m_auth_list.at( i )->share().isEmpty() ||
               QString::compare( m_auth_list.at( i )->share().toUpper(),
               authInfo->share().toUpper() ) == 0) )
          {
            known = m_auth_list[i];

            if ( authInfo->share().isEmpty() == known->share().isEmpty() )
            {
              // Exact match. Stop here.
              break;
            }
            else
            {
              // Continue until we found the exact match or reached
              // the end of the list.
              continue;
            }
          }
          else
          {
            continue;
          }
        }

        if ( known )
        {
          // Overwrite the previous data. The host name and the share name
          // does not need to be replaced.
          known->setWorkgroup( authInfo->workgroup() );
          known->setUser( authInfo->user() );
          known->setPassword( authInfo->password() );
        }
        else
        {
          m_auth_list.append( new Smb4KAuthInfo( *authInfo ) );
        }

        success = true;
      }
      else
      {
        // Fill the temporary authentication information into
        // m_temp_auth.
        if ( !m_temp_auth )
        {
          m_temp_auth = new Smb4KAuthInfo( *authInfo );
        }
        else
        {
          // Do nothing
        }

        success = true;
      }
    }

#ifdef __FreeBSD__

    writeToSMBConfFile( authInfo );

#endif
  }
  else
  {
    return success;  // false
  }

  return success;
}


bool Smb4KPasswordHandler::readDefaultAuth( Smb4KAuthInfo *authInfo )
{
  bool success = false;

  if ( m_can_read_auth_info && authInfo )
  {
    if ( walletIsOpen() && Smb4KSettings::useWallet() )
    {
      // Read the default authentication information.
      QMap<QString,QString> map;

      m_wallet->readMap( "DEFAULT_LOGIN", map );

      if ( !map.isEmpty() )
      {
        authInfo->setUser( map["Login"] );
        authInfo->setPassword( map["Password"] );

        success = true;
      }
      else
      {
        return success;  // false
      }
    }
    else
    {
      if ( !Smb4KSettings::useWallet() && walletIsOpen() )
      {
        delete m_wallet;
        m_wallet = NULL;
      }
      else
      {
        // Do nothing
      }

      return success;  // false
    }
  }
  else
  {
    return success;  // false
  }

  return success;
}


bool Smb4KPasswordHandler::writeDefaultAuth( Smb4KAuthInfo *authInfo )
{
  bool success = false;

  if ( m_can_read_auth_info && authInfo )
  {
    if ( walletIsOpen() && Smb4KSettings::useWallet() )
    {
      QMap<QString,QString> map;
      map["Login"]    = authInfo->user();
      map["Password"] = authInfo->password();

      m_wallet->writeMap( "DEFAULT_LOGIN", map );

      if ( m_wallet->sync() == 0 )
      {
        success = true;
      }
      else
      {
        return success;  // false
      }
    }
    else
    {
      if ( !Smb4KSettings::useWallet() && walletIsOpen() )
      {
        delete m_wallet;
        m_wallet = NULL;
      }
      else
      {
        // Do nothing
      }

      return success;  // false
    }
  }
  else
  {
    return success;  // false
  }

  return success;
}


#ifdef __FreeBSD__

void Smb4KPasswordHandler::writeToSMBConfFile( Smb4KAuthInfo *authInfo )
{
  m_nsmbrc_auth = *authInfo;

  m_proc->setShellCommand( QString( "smbutil crypt %1" )
          .arg( QString::fromLocal8Bit( m_nsmbrc_auth.password() ) ) );

  m_proc->setOutputChannelMode( KProcess::SeparateChannels );
  m_proc->start();
}

#endif


/////////////////////////////////////////////////////////////////////////////
//  SLOT IMPLEMENTATIONS
/////////////////////////////////////////////////////////////////////////////


void Smb4KPasswordHandler::slotWalletOpened( bool success )
{
  if ( success )
  {
    if ( !m_wallet->hasFolder( "Smb4K" ) )
    {
      m_wallet->createFolder( "Smb4K" );
      m_wallet->setFolder( "Smb4K" );
    }
    else
    {
      m_wallet->setFolder( "Smb4K" );
      convert_old_entries();
    }

    m_wallet_support_disabled = false;
  }
  else
  {
    Smb4KCoreMessage::information( INFO_OPENING_WALLET_FAILED, KWallet::Wallet::NetworkWallet(), QString() );

    delete m_wallet;
    m_wallet = NULL;

    m_wallet_support_disabled = true;
  }

  m_can_read_auth_info = true;
  emit readyReadAuthInfo();
}


void Smb4KPasswordHandler::slotGetPassword( const QString &username )
{
  if ( m_dlg && m_auth )
  {
    // We need to re-read the auth data here, because
    // of the homes shares:
    Smb4KAuthInfo authInfo( m_auth->workgroup().toUpper(),
                            m_auth->host().toUpper(), username );

    readAuth( &authInfo );

    KLineEdit *lineEdit = m_dlg->findChild<KLineEdit *>( "AskPassPasswordEdit" );
    lineEdit->setText( authInfo.password() );

    m_auth->setShare( username );
  }
}


void Smb4KPasswordHandler::slotEnableOKButton( const QString &text )
{
  if ( m_dlg )
  {
    m_dlg->enableButtonOk( !text.isEmpty() );
  }
}

#ifdef __FreeBSD__
void Smb4KPasswordHandler::slotProcessFinished( int exitCode, QProcess::ExitStatus exitStatus )
{
  if ( exitCode == 0 && exitStatus == QProcess::NormalExit )
  {
    // Read from stdout:
    QString stdout_output = QString::fromLocal8Bit( m_proc->readAllStandardOutput(), -1 );

    QString pass = stdout_output.remove( "\n" ).trimmed();

    QDir::setCurrent( QDir::homePath() );

    QFile file( ".nsmbrc" );

    QStringList contents;

    if ( file.exists() )
    {
      if ( file.open( QIODevice::ReadOnly | QIODevice::Text ) )
      {
        QTextStream ts( &file );
        // Note: With Qt 4.3 this seems to be obsolete, but we'll
        // keep it for now.
        ts.setCodec( QTextCodec::codecForLocale() );

        QString line;

        while ( 1 )
        {
          line = ts.readLine( 0 );

          if ( !line.isNull() )
          {
            contents.append( line );
          }
          else
          {
            break;
          }
        }

        file.close();
      }
    }

    // Check if the [default] section is present.
    if ( !contents.contains( "[default]", Qt::CaseInsensitive ) )
    {
      QMap<QString,QString> map = Smb4KSambaOptionsHandler::self()->globalSambaOptions();

      QStringList global_section;

      // Insert default section
      global_section.append( "[default]" );

      // Workgroup
      if ( Smb4KSettings::domainName().isEmpty() )
      {
        if ( !map["Workgroup"].isEmpty() )
        {
          global_section.append( "workgroup="+map["workgroup"] );
        }
        else
        {
          // Do nothing
        }
      }
      else
      {
        global_section.append( "workgroup="+Smb4KSettings::domainName() );
      }

      // WINS server
      if ( !map["wins server"].isEmpty() )
      {
        global_section.append( "nbns="+map["wins server"] );
      }
      else
      {
        // Do nothing
      }

      // Character sets
      if ( Smb4KSettings::clientCharset() != Smb4KSettings::EnumClientCharset::default_charset &&
           Smb4KSettings::serverCodepage() != Smb4KSettings::EnumServerCodepage::default_codepage )
      {
        global_section.append( "charsets="+Smb4KSettings::self()->clientCharsetItem()->label()
                                +":"+Smb4KSettings::self()->serverCodepageItem()->label() );
      }
      else
      {
        // Do nothing
      }

      global_section.append( "" );

      while ( !global_section.isEmpty() )
      {
        contents.prepend( global_section.takeLast() );
      }
    }
    else
    {
      // Do nothing
    }

    QString section;

    if ( m_nsmbrc_auth.share().isEmpty() )
    {
      section.append( QString( "[%1:%2]" ).arg( m_nsmbrc_auth.host().toUpper(),
                      (m_nsmbrc_auth.user().isEmpty() ?
                      "GUEST" :
                      QString::fromLocal8Bit( m_nsmbrc_auth.user() ).toUpper()) ) );
    }
    else
    {
      section.append( QString( "[%1:%2:%3]" ).arg( m_nsmbrc_auth.host().toUpper(),
                      (m_nsmbrc_auth.user().isEmpty() ?
                      "GUEST" :
                      QString::fromLocal8Bit( m_nsmbrc_auth.user() ).toUpper()),
                      m_nsmbrc_auth.share().toUpper() ) );
    }

    int index = contents.indexOf( section );

    if ( index == -1 )
    {
      if ( !contents.last().isEmpty() )
      {
        contents.append( "" );
      }
      else
      {
        // Do nothing
      }

      contents.append( section );
      contents.append( "password="+pass );

      if ( !m_nsmbrc_auth.workgroup().isEmpty() )
      {
        contents.append( "workgroup="+m_nsmbrc_auth.workgroup() );
      }
      else
      {
        // Do nothing
      }

      kDebug() << "Include custom character sets..." << endl;
    }
    else
    {
      for ( int i = ++index; i < contents.size(); ++i )
      {
        if ( contents.at( i ).startsWith( "password=" ) )
        {
          contents[i] = "password="+pass;

          break;
        }
        else if ( contents.at( i ).startsWith( "[" ) )
        {
          break;
        }
        else
        {
          continue;
        }
      }
    }

    QDir::setCurrent( QDir::homePath() );

    if  ( file.open( QIODevice::WriteOnly | QIODevice::Text ) )
    {
      QTextStream ts( &file );
      // Note: With Qt 4.3 this seems to be obsolete, we'll keep
      // it for now.
      ts.setCodec( QTextCodec::codecForLocale() );

      for ( int i = 0; i < contents.size(); ++i )
      {
        ts << contents.at( i ) << endl;
      }

      file.close();
    }
    else
    {
      Smb4KCoreMessage::error( ERROR_OPENING_FILE, "~/.nsmbrc", QString() );

      return;
    }

    // Get minimal security: Fix permissions.
    QString path = QDir::homePath()+"/"+file.fileName();

    int fd;

    if ( (fd = open( path.toLocal8Bit(), O_WRONLY )) == -1 )
    {
      int err_no = errno;
      Smb4KCoreMessage::error( ERROR_OPENING_FILE, path, strerror( err_no ) );

      return;
    }
    else
    {
      if ( fchmod( fd, 00600 ) == -1 )
      {
        // FIXME: Do we need to emit an error here at all?
      }
    }
  }
  else
  {
    // Something went wrong. Throw an error.
    if ( m_process_error != -1 )
    {
      Smb4KCoreMessage::processError( ERROR_PROCESS_ERROR, m_process_error );
    }
    else
    {
      Smb4KCoreMessage::processError( ERROR_PROCESS_EXIT, m_process_error );
    }
  }

  m_proc->clearProgram();
  m_process_error = (QProcess::ProcessError)(-1);
#else
void Smb4KPasswordHandler::slotProcessFinished( int, QProcess::ExitStatus )
{
#endif
}


#ifdef __FreeBSD__
void Smb4KPasswordHandler::slotProcessError( QProcess::ProcessError errorCode )
{
  m_process_error = errorCode;
#else
void Smb4KPasswordHandler::slotProcessError( QProcess::ProcessError )
{
#endif
}

#include "smb4kpasswordhandler.moc"
