/***************************************************************************
    smb4kpasswordhandler  -  This class handles the passwords for Smb4K.
                             -------------------
    begin                : So Jan 16 2005
    copyright            : (C) 2005-2008 by Alexander Reinholdt
    email                : dustpuppy@users.berlios.de
 ***************************************************************************/

/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful, but   *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,   *
 *   MA  02111-1307 USA                                                    *
 ***************************************************************************/

#ifndef SMB4KPASSWORDHANDLER_H
#define SMB4KPASSWORDHANDLER_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// Qt includes
#include <QObject>
#include <QList>
#include <QString>

// KDE includes
#include <kwallet.h>
#include <kdialog.h>
#include <kprocess.h>
#include <kdemacros.h>

// forward declarations
#ifndef __FreeBSD__
class Smb4KAuthInfo;
#endif

// application specific includes
#include "smb4khomesshareshandler.h"
#ifdef __FreeBSD__
#include "smb4ksambaoptionshandler.h"
#include "smb4kauthinfo.h"
#endif



/**
 * This class handles the passwords used by Smb4K to authenticate to
 * the network shares.
 *
 * @author Alexander Reinholdt <dustpuppy@users.berlios.de>
 */

class KDE_EXPORT Smb4KPasswordHandler : public QObject
{
  Q_OBJECT

  friend class Smb4KPasswordHandlerPrivate;

  public:
    /**
     * Returns a static pointer to this class.
     */
    static Smb4KPasswordHandler *self();

    /**
     * This enumeration destermines which destriptive label the askpass
     * dialog should carry.
     */
    enum Description{ NewData,
                      AccessDenied,
                      BadPassword,
                      PermissionDenied,
                      AuthenticationError,
                      LogonFailure,
                      None };

    /**
     * Opens the askpass dialog.
     *
     * If the password handler was not initialized through the init() function before this
     * dialog is invoked, the initialization will be done here. Since the wallet will be
     * opened synchronously, this may have some drawbacks.
     *
     * @param workgroup     The workgroup where the share is located in (may be empty).
     *
     * @param host          The host the share belongs to.
     *
     * @param name          The name of the share.
     *
     * @param desc          Determines which descriptive label the askpass dialog should carry.
     *                      Values are taken from the Smb4KPasswordHandler::AskPass::Description
     *                      enumeration.
     *
     * @param parent        The parent of this dialog. If the parent is 0, an appropriate parent
     *                      will automatically be chosen.
     *
     * @returns             TRUE if a new password has been supplied and FALSE otherwise.
     */
    bool askpass( const QString &workgroup,
                  const QString &host,
                  const QString &share,
                  Description desc,
                  QWidget *parent = 0 );

    /**
     * This function looks for authentication information. It fills @p authInfo with it and
     * returns TRUE upon success.
     *
     * You must run init() (or the askpass dialog) before you can access the authentication
     * information. If the password handler is not initialized, this function will always
     * return FALSE. No authentication information will be provided in that case.
     *
     * @param authInfo      The authentication information object.
     *
     * @returns TRUE upon success and false otherwise.
     */
    bool readAuth( Smb4KAuthInfo *authInfo );

    /**
     * Write the provided authentication information to the wallet or the internal temporary
     * list. The function returns TRUE upon success.
     *
     * You must run init() (or the askpass dialog) before you can access the authentication
     * information. If the password handler is not initialized, this function will always
     * return FALSE. No authentication information will written in that case.
     *
     * @param authInfo      The authentication information object.
     *
     * @returns TRUE upon success and false otherwise.
     */
    bool writeAuth( Smb4KAuthInfo *authInfo );

    /**
     * This function looks for the default authentication information. It fills @p authInfo
     * with it and returns TRUE upon success.
     *
     * You must run init() (or the askpass dialog) before you can access the authentication
     * information. If the password handler is not initialized, this function will always
     * return FALSE. No authentication information will be provided in that case.
     *
     * @param authInfo      The authentication information object.
     *
     * @returns TRUE upon success and false otherwise.
     */
    bool readDefaultAuth( Smb4KAuthInfo *authInfo );

    /**
     * Write the provided default authentication information to the wallet. The function
     * returns TRUE upon success.
     *
     * You must run init() (or the askpass dialog) before you can access the authentication
     * information. If the password handler is not initialized, this function will always
     * return FALSE. No authentication information will written in that case.
     *
     * @param authInfo      The authentication information object.
     *
     * @returns TRUE upon success and false otherwise.
     */
    bool writeDefaultAuth( Smb4KAuthInfo *authInfo );

    /**
     * This function initializes the password handler. It tries to open the wallet
     * either asynchronously or synchronously, depending on the value of @p async.
     * If you choose to open the wallet asynchronously (the default), it cannot be
     * accessed immediately but only after the opening process finished. Connect to
     * the readyReadAuthInfo() signal and use canReadAuthInfo() to find out if you
     * can retrieve authentication information. If the opening of the wallet failed,
     * walletIsOpen() and walletSupportIsDisabled() will additionally return FALSE.
     *
     * Optionally you can provide a parent widget @p parent that can be used if
     * init() is run from a GUI. If @p parent is null, kapp->activeWindow() is used.
     *
     * @param parent        The optional parent widget
     *
     * @param async         Open the wallet asynchronously.
     */
    void init( QWidget *parent = 0,
               bool async = true );

    /**
     * This function returns TRUE if a digital wallet exists and is open. Otherwise
     * it returns FALSE.
     *
     * If you only want to check if the password handler will read and return authentication data
     * upon request, use the canReadAuthInfo() function or connect to the readyReadAuthInfo()
     * signal instead.
     *
     * @returns TRUE if the wallet exists and is open.
     */
    bool walletIsOpen() { return (m_wallet && m_wallet->isOpen()); }

    /**
     * Check if you can read authentication information. This function will return TRUE after
     * init() was run and the password handler is in a state where you can retrieve authentication
     * information upon request. It does not matter if the wallet is open or not. If you need to
     * access the wallet, you should use walletIsOpen() in addition.
     *
     * @returns TRUE if the password handler is in a state where you can retrieve authentication
     * information upon request.
     */
    bool canReadAuthInfo() { return m_can_read_auth_info; }

    /**
     * This function returns TRUE if the wallet support has been temporarily diabled, that means
     * the user did not supply the correct password when opening the wallet.
     *
     * @returns TRUE if the wallet support is temporarily disabled.
     */
    bool walletSupportIsDisabled() { return m_wallet_support_disabled; }

  signals:
    /**
     * This signal is emitted when the password handler entered a state where you can retieve
     * authentication information upon request.
     *
     * Please note that this signal is always emitted after init() was run, so you have to test
     * by using walletIsOpen() if the digital wallet is open and you can access it.
     */
    void readyReadAuthInfo();

  protected slots:
    /**
     * This slot is called when an attempt was made to open the requested
     * wallet.
     *
     * @param success       TRUE when the opening was successful.
     */
    void slotWalletOpened( bool success );

    /**
     * Get the password for a specific share and user name.
     */
    void slotGetPassword( const QString &username );

    /**
     * This slot is used to enable the OK button of the askpass
     * dialog when text has been entered in the text boxes and/or the
     * combo box.
     *
     * @param text          The text that has been entered.
     */
    void slotEnableOKButton( const QString &text );

    /**
     * This slot is called, when the process exited.
     *
     * @param exitCode           The exit code of the process
     *
     * @param exitStatus         The exit status of the process (@see QProcess::ExitStatus)
     */
    void slotProcessFinished( int exitCode, QProcess::ExitStatus exitStatus );

    /**
     * This slot is invoked when an error occurred with the process.
     *
     * @param errorCode       The error code returned by the process
     */
    void slotProcessError( QProcess::ProcessError errorCode );

  private:
    /**
     * The constructor.
     */
    Smb4KPasswordHandler();

    /**
     * The destructor.
     */
    ~Smb4KPasswordHandler();

    /**
     * This function is used to convert old wallet entries to the new
     * map based ones.
     */
    void convert_old_entries();

    /**
     * The KWallet object.
     */
    KWallet::Wallet *m_wallet;

    /**
     * The Smb4KAuthInfo object. For use with the askpass dialog.
     */
    Smb4KAuthInfo *m_auth;

    /**
     * The askpass dialog.
     */
    KDialog *m_dlg;

    /**
     * This list holds the authentication data the user supplied if he/she
     * does not want to use KWallet.
     */
    QList<Smb4KAuthInfo *> m_auth_list;

    /**
     * If the user neither wants to store the passwords in a wallet nor in a
     * temporary list, the authentication data received from the askpass dialog
     * will the stored in this object.
     */
    Smb4KAuthInfo *m_temp_auth;

    /**
     * Is the wallet support diabled?
     */
    bool m_wallet_support_disabled;

    /**
     * Can we read authentication information?
     */
    bool m_can_read_auth_info;

#ifdef __FreeBSD__

    /**
     * The KProcess object
     */
    KProcess *m_proc;

    /**
     * FreeBSD specific: This function writes authentication information
     * to the ~/.nsmbrc file, which is used by mount_smbfs.
     *
     * @param authInfo      The authentication information that should be written
     *                      to ~/.nsmbrc file.
     */
    void writeToSMBConfFile( Smb4KAuthInfo *authInfo );

    /**
     * FreeBSD specific: Holds the authentication information that should
     * be written to ~/.nsmbrc.
     */
    Smb4KAuthInfo m_nsmbrc_auth;

    /**
     * The process error code
     */
    QProcess::ProcessError m_process_error;
#endif
};

#endif
